onew c {1 9}
oaddev {ctl  c 70 1}
oaddev {ctl  c 71 2}
oaddev {xpc  c 255 64}
oaddev {xctl c 7 1023}
oaddev {nrpn c 1 10}
oaddev {rpn  c 2 20}
oaddev {nrpn c 3 30}
oaddev {nrpn c 4 40}
g 0; sel 0; ct nil; ci nil; co nil
