/*
 * Copyright (c) 2003-2010 Alexandre Ratchov <alex@caoua.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdio.h>
#include <unistd.h>

#include "utils.h"
#include "str.h"
#include "cons.h"
#include "ev.h"
#include "mux.h"
#include "track.h"
#include "frame.h"
#include "song.h"
#include "user.h"
#include "filt.h"
#include "mididev.h"
#include "defs.h"
#include "sysex.h"
#include "textio.h"

int
main(int argc, char **argv)
{
	int ch;
	unsigned exitcode;

	while ((ch = getopt(argc, argv, "bv")) != -1) {
		switch (ch) {
		case 'b':
			user_flag_batch = 1;
			break;
		case 'v':
			user_flag_verb = 1;
			break;
		default:
			goto err;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc >= 1) {
	err:
		fputs("usage: midish [-bv]\n", stderr);
		return 0;
	}

	exitcode = user_mainloop();

	return exitcode ? 0 : 1;
}
