/*
 * Copyright (c) 2003-2010 Alexandre Ratchov <alex@caoua.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef MIDISH_CONS_H
#define MIDISH_CONS_H

#include "tty.h"

void cons_init(struct el_ops *, void *);
void cons_done(void);
void cons_putpos(unsigned, unsigned, unsigned);
void cons_puttag(char *);
void cons_ready(void);

void cons_err(char *);
void cons_errs(char *, char *);
void cons_erru(unsigned long, char *);
void cons_errss(char *, char *, char *);
void cons_errsu(char *, unsigned long, char *);
void cons_erruu(unsigned long, unsigned long, char *);

extern int cons_isatty;

#endif /* MIDISH_CONS_H */
