/*
 * Copyright (c) 2003-2010 Alexandre Ratchov <alex@caoua.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef MIDISH_CONV_H
#define MIDISH_CONV_H

#include "ev.h"

#define CONV_NUMREV 4

/*
 * constants to define a set of packed events
 */
#define CONV_XPC	(1 << EV_XPC)
#define CONV_NRPN	(1 << EV_NRPN)
#define CONV_RPN	(1 << EV_RPN)

struct ev;
struct statelist;

unsigned conv_packev(struct statelist *, unsigned, unsigned,
    struct ev *, struct ev *);
unsigned conv_unpackev(struct statelist *, unsigned, unsigned,
    struct ev *, struct ev *);

#endif /* MIDISH_CONV_H */
